

// (c) COPYRIGHT URI/MIT 1997-98
// Please read the full copyright statement in the file COPYRIGHT.  
//
// Authors: reza (Reza Nekovei)

// This file contains functions which read the variables and their attributes
// from a netcdf file and build the in-memeory DAS. These functions form the
// core of the server-side software necessary to extract the DAS from a
// netcdf data file.
//
// ReZa 6/23/97

// $Log: ffdas.cc,v $
// Revision 1.11.2.1  1999/08/28 01:25:53  jimg
// Modified the code which reads header variables and loads them into the DAS in
// read_attributes(). When the second FF db_ask(...) call returns an error this
// code should assume that there are no header variables, not that a fatal error
// was detected. This assumes that the previous db_ask(...) call was successful.
// Since there are no header variables when the second db_ask(...) call returns
// an error, make sure not to access the pinfo_list data structure.
//
// Revision 1.11  1999/07/22 21:28:10  jimg
// Merged changes from the release-3-0-2 branch
//
// Revision 1.10.2.1  1999/07/21 20:27:08  dan
// Added Attribute handling for FreeForm File_Headers.  FreeForm
// header variables defined in the format-descriptor file will be
// added the the FF_GLOBAL DAS container.  File_Headers are commonly
// used by FreeForm to provide metadata concerning the data contained
// in the dataset.
//
// Revision 1.10  1999/05/27 17:02:23  jimg
// Merge with alpha-3-0-0
//
// Revision 1.9.2.1  1999/05/20 21:43:51  edavis
// Fix spelling of COPYRIGHT and remove some #if 0 stuff.
//
// Revision 1.9  1999/05/04 02:55:38  jimg
// Merge with no-gnu
//
// Revision 1.8  1999/03/26 20:03:32  jimg
// Added support for the Int16, UInt16 and Float32 datatypes
//
// Revision 1.7.12.1  1999/05/01 04:40:30  brent
// converted old String.h to the new std C++ <string> code
//
// Revision 1.7  1998/08/31 04:06:13  reza
// Added String support.
// Fixed data alignment problem (64-bit Architectures).
// Removed Warnings and added a check for file existence.
// Updated FFND to fix a bug in stride.
//
// Revision 1.6  1998/08/14 18:26:33  reza
// Removed embedded double quotes in attributes.
//
// Revision 1.5  1998/08/13 20:23:45  jimg
// Removed the attribute about file info.
//
// Revision 1.4  1998/08/12 21:21:14  jimg
// Massive changes from Reza. Compatible with the new FFND library
//
// Revision 1.3  1998/04/21 17:14:09  jimg
// Fixes for warnings, etc
//
// Revision 1.2  1998/04/16 18:11:24  jimg
// Sequence support added by Reza
//
//

#include "config_ff.h"

static char rcsid[] not_used = {"$Id: ffdas.cc,v 1.11.2.1 1999/08/28 01:25:53 jimg Exp $"};

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <iostream.h>
#include <string>

#include "cgi_util.h"
#include "DAS.h"
// Hack. Header files from the WWW library (which are included from Connect.h
// within cgi_util.h) also define BOOLEAN. Suppressing the definition avoids
// a warning (g++ 2.8.1) or error (egcs). 3/26/99 jhrg
#define _BOOLEAN_DEFINED	
#include "FreeForm.h"
#include "util_ff.h"
#include "freeform.h"

int StrLens[MaxStr]; // List of string lengths

// Used by ErrMsgT
static char Msgt[255];

// reads the attributes and store their names and values in the attribute
// table. 
//
// Returns: false if an error was detected reading from the freeform format
// file, true otherwise. 

bool
read_attributes(string filename, AttrTable *at, string *err_msg)
{
    int error = 0;
    size_t pos;
    FF_BUFSIZE_PTR bufsize = NULL;
    DATA_BIN_PTR dbin = NULL;
    FF_STD_ARGS_PTR SetUps = NULL;  

    if (!file_exist(filename.c_str())) {
	sprintf(Msgt, "ff_das: Could not open %s", filename.c_str());
	ErrMsgT(Msgt);  
	*err_msg = (string)"\"" + Msgt + " \"";
	return false;
    }
    
    SetUps = ff_create_std_args();
    if (!SetUps) {
	sprintf(Msgt, "ff_das: Insufficient memory -- free more memory and try again");
	ErrMsgT(Msgt);  
	*err_msg = (string)"\"" + Msgt + " \"";
	return false;
    }
    
    /** set the structure values to create the FreeForm DB**/
    SetUps->user.is_stdin_redirected = 0;

    SetUps->input_file = new char[filename.length() + 1];
    strcpy(SetUps->input_file, filename.c_str());

    string iff = find_ancillary_file(filename);
    SetUps->input_format_file = new char[iff.length() + 1];
    strcpy(SetUps->input_format_file, iff.c_str()); // strcpy needs the /0
    SetUps->output_file = NULL;

    error = SetDodsDB(SetUps, &dbin, Msgt);
    if (error && error < ERR_WARNING_ONLY) {
	db_destroy(dbin);
	ErrMsgT(Msgt);  
	*err_msg = "\"" + (string)Msgt + " \"";
	return false;
    }

    error = db_ask(dbin,DBASK_FORMAT_SUMMARY,FFF_INPUT, &bufsize);
    if (error) {
	sprintf(Msgt, "ff_das: db_ask cannot get summary information");
	ErrMsgT(Msgt);  
	*err_msg = "\"" + (string)Msgt + " \"";
	return false;
    }

    at->append_attr("Server", "STRING", "\"DODS FreeFrom based on FFND release "FFND_LIB_VER"\"");

    //fix the format of the info. string
    string fmt_info = bufsize->buffer;
    while ((pos = fmt_info.find('"')) < fmt_info.size()) 
	fmt_info.replace(pos, 1, '`');
    while ((pos = fmt_info.find('\n')) < fmt_info.size()) 
	fmt_info.replace(pos, 1, ' ');
    fmt_info = "\"" + fmt_info + "\"";
    at->append_attr("Native_file", "STRING", fmt_info);

    PROCESS_INFO_LIST pinfo_list = NULL;
    PROCESS_INFO_PTR hd_pinfo = NULL;

    char text[256];

    error = db_ask(dbin, DBASK_PROCESS_INFO, FFF_INPUT | FFF_FILE | FFF_HEADER, &pinfo_list);
#if 0
    if (error) {
	sprintf(Msgt, "ff_das: db_ask cannot get header information");
	ErrMsgT(Msgt);  
	*err_msg = "\"" + (string)Msgt + " \"";
	return false;
    }
#endif

    if (!error) {
	pinfo_list = dll_first(pinfo_list);
	hd_pinfo = ((PROCESS_INFO_PTR)(pinfo_list)->data.u.pi);
	if (hd_pinfo) {
      
	    VARIABLE_LIST vlist = NULL;
	    VARIABLE_PTR var = NULL;
      
	    vlist = FFV_FIRST_VARIABLE(PINFO_FORMAT(hd_pinfo));
	    var = ((VARIABLE_PTR)(vlist)->data.u.var);

	    while (var) {
		if (IS_EOL(var)) {
		    vlist = (vlist)->next;
		    var = ((VARIABLE_PTR)(vlist)->data.u.var);
		
		    continue;
		}

		switch (FFV_DATA_TYPE(var)) {

		  case FFV_TEXT:
		    nt_ask(dbin, FFF_FILE | FFF_HEADER, var->name, FFV_TEXT, text);
		    at->append_attr(var->name, "STRING", text);
		    break;

		  case FFV_INT8:
		    unsigned char int8;
		    nt_ask(dbin, FFF_FILE | FFF_HEADER, var->name, FFV_INT8, &int8);
		    sprintf(text,"%d",int8);
		    at->append_attr(var->name, "BYTE", text);
		    break;

		  case FFV_INT16:
		    short int16;
		    nt_ask(dbin, FFF_FILE | FFF_HEADER, var->name, FFV_INT16, &int16);
		    sprintf(text,"%d",int16);
		    at->append_attr(var->name, "INT16", text);
		    break;

		  case FFV_INT32:
		    int int32;
		    nt_ask(dbin, FFF_FILE | FFF_HEADER, var->name, FFV_INT32, &int32);
		    sprintf(text,"%d",int32);
		    at->append_attr(var->name, "INT32", text);
		    break;

		  case FFV_INT64:
		    long int64;
		    nt_ask(dbin, FFF_FILE | FFF_HEADER, var->name, FFV_INT64, &int64);
		    sprintf(text,"%ld",int64);
		    at->append_attr(var->name, "INT32", text);
		    break;

		  case FFV_UINT8:
		    unsigned char uint8;
		    nt_ask(dbin, FFF_FILE | FFF_HEADER, var->name, FFV_UINT8, &uint8);
		    sprintf(text,"%d",uint8);
		    at->append_attr(var->name, "BYTE", text);
		    break;

		  case FFV_UINT16:
		    unsigned short uint16;
		    nt_ask(dbin, FFF_FILE | FFF_HEADER, var->name, FFV_UINT16, &uint16);
		    sprintf(text,"%d",uint16);
		    at->append_attr(var->name, "UINT16", text);
		    break;

		  case FFV_UINT32:
		    unsigned int uint32;
		    nt_ask(dbin, FFF_FILE | FFF_HEADER, var->name, FFV_UINT32, &uint32);
		    sprintf(text,"%u",uint32);
		    at->append_attr(var->name, "UINT32", text);
		    break;

		  case FFV_UINT64:
		    unsigned long uint64;
		    nt_ask(dbin, FFF_FILE | FFF_HEADER, var->name, FFV_UINT64, &uint64);
		    sprintf(text,"%lu",uint64);
		    at->append_attr(var->name, "UINT32", text);
		    break;

		  case FFV_FLOAT32:
		    float float32;
		    nt_ask(dbin, FFF_FILE | FFF_HEADER, var->name, FFV_FLOAT32, &float32);
		    sprintf(text,"%f",float32);
		    at->append_attr(var->name, "FLOAT32", text);
		    break;

		  case FFV_FLOAT64:
		    double float64;
		    nt_ask(dbin, FFF_FILE | FFF_HEADER, var->name, FFV_FLOAT64, &float64);
		    sprintf(text,"%f",float64);
		    at->append_attr(var->name, "FLOAT64", text);
		    break;

		  case FFV_ENOTE:
		    double enote;
		    nt_ask(dbin, FFF_FILE | FFF_HEADER, var->name, FFV_ENOTE, &enote);
		    sprintf(text,"%e",enote);
		    at->append_attr(var->name, "FLOAT64", text);
		    break;

		  default:
		    assert("Unknown FreeForm type!" && false);
		    break;
		}           
		vlist = (vlist)->next;
		var = ((VARIABLE_PTR)(vlist)->data.u.var);
	    }
	}
    }

    return true;
}

// Given a reference to an instance of class DAS and a filename that refers
// to a freefrom file, read the format file and extract the existing
// attributes and add them to the instance of DAS.
//
// Returns: false if an error accessing the file was detected, true
// otherwise. 

bool
get_attributes(DAS &das, string filename, string *error)
{

    AttrTable *attr_table_ptr;
  
    // global attributes (no variable attributes in freefrom)
    attr_table_ptr = das.add_table("FF_GLOBAL", new AttrTable);
    if (!read_attributes(filename, attr_table_ptr, error))
	return false;

  
    return true;
}

#ifdef TEST

int
main(int argc, char *argv[])
{
    DAS das;

    if(!get_attributes(&das, (string)argv[1], ""))
	abort();

    das.print();
}

#endif



